/* make-unicode-tables.c
 *
 * Creates lookup tables for the unicode characters. */

/* Copyright (C) 2002 Niels Mller
 *  
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>

#include <stdio.h>
#include <stdlib.h>

#include <string.h>

#include "unidata.h"

/* Perhaps too large for the stack. Also makes sure it is
 * initialized with all zeroes */
static struct unidata_entry unidata[UNICODE_MAX_CHAR];

#define MAX_LINE 500

static int
read_unidata(FILE *f)
{
  unsigned char line[MAX_LINE+1];
  wchar_t start = 0;
  
  line[MAX_LINE] = '\0';

  while (fgets(line, MAX_LINE, f))
  {
    struct unidata_entry entry;
    if (unidata_parse_entry(&entry, line))
    {
      assert(entry.code >= 0);
      if (entry.code >= UNICODE_MAX_CHAR)
	/* Ignore */
	continue;
#if 0
      fprintf(stderr, "Parsed %04lx\n", (long) entry.code);
#endif
      
      if (entry.start)
	start = entry.code;
      else
      {
	if (!entry.end)
	  unidata[entry.code] = entry;
	else
	{
	  wchar_t c;
	  if (!start)
	  {
	    fprintf(stderr, "End range without matching start: %s\n",
		    entry.name);
	    return 0;
	  }
	  for (c = start; c <= entry.end; c++)
	  {
	    unidata[c] = entry;
	    unidata[c].code = c;
	  }
	}
	
	start = 0;
      }
    } else {
      fprintf(stderr, "Failed to parse line %s\n", line);
      return 0;
    }
  }
  return 1;
}

struct table
{
  unsigned subtable[UNICODE_MAX_MASTER][0x100];
  unsigned mastertable[UNICODE_MAX_MASTER];
  int n;
};

static void
table_init(struct table *table)
{
  table->n = 0;
}

static void
table_add_subtable(struct table *table,
		   unsigned master,
		   const unsigned *subtable)
{
  unsigned i;

  assert(master < UNICODE_MAX_MASTER);
  assert(master >= table->n);
  for (i = 0; i<table->n; i++)
  {
    unsigned j;
    for (j = 0; j<0x100; j++)
      if (table->subtable[i][j] != subtable[j])
	goto next_subtable;

    /* Same as subtable i */
    table->mastertable[master] = i;
    return;

  next_subtable:
    ;
  }

  /* A brand new subtable */
  memcpy(table->subtable[table->n], subtable,
	 sizeof(table->subtable[table->n]));
  table->mastertable[master] = table->n++;
}

static void
table_write(struct table *table, const char *type, const char *name, FILE *f)
{
  unsigned i;
  for (i = 0; i<table->n; i++)
  {
    unsigned int j;
    
    fprintf(f, "static const %s %s_%d[0x100] = {", type, name, i);
    for (j = 0; j<0x100; j++)
    {
      if (!(j % 16))
	fprintf(f, "\n  ");
      fprintf(f, "%d,", table->subtable[i][j]);
    }
    fprintf(f, "\n};\n\n");
  }

  fprintf(f, "const %s *%s[%d] = {", type, name, UNICODE_MAX_MASTER);
  for (i = 0; i<UNICODE_MAX_MASTER; i++)
  {
    if (!(i%4))
      fprintf(f, "\n  ");
    fprintf(f, "%s_%d, ", name, table->mastertable[i]);
  }
  fprintf(f, "\n};\n\n");
}

static void
write_combining_table(void)
{
  struct table table;
  unsigned subtable[0x100];
  unsigned master;

  table_init(&table);
  
  for (master = 0; master < UNICODE_MAX_MASTER; master++)
  {
    unsigned i;
    
    /* Build one 256-entry table */
    for (i = 0; i<0x100; i++)
    {
      wchar_t c = master * 0x100 + i; 
      subtable[i] = (c < UNICODE_MAX_CHAR) ? unidata[c].combining_class : 0;
    }

    table_add_subtable(&table, master, subtable);
  }

  table_write(&table, "unsigned char", "unicode_cclass", stdout);
}

int
main(int argc, char **argv)
{
  if (!read_unidata(stdin))
    return EXIT_FAILURE;

  write_combining_table();
  
  return EXIT_SUCCESS;
}
