/* unidata.c
 *
 * Functions for parsing the unidata text file. */

/* Copyright (C) 2002 Niels Mller
 *  
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>

#include "unidata.h"

struct buffer
{
  const char *current;
  unsigned length;
  const char *next;
};

static void
buffer_init(struct buffer *buffer, const unsigned char *line)
{
  buffer->current = NULL;
  buffer->next = line;
}

static int
buffer_next(struct buffer *buffer)
{
  buffer->current = buffer->next;
  if (!buffer->current)
    return 0;
  buffer->next = strchr(buffer->current, ';');
  if (!buffer->next)
    return 0;

  buffer->length = buffer->next - buffer->current;
  buffer->next++;

  return 1;
}

static int
parse_hex(const unsigned char *digits, unsigned length, wchar_t *value)
{
  wchar_t c;
  unsigned i;

  const signed char hex_digits[0x100] =
    {
      -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
      -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
      -1,-1,-1,-2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
      0, 1, 2, 3, 4, 5, 6, 7, 8, 9,-1,-1,-1,-1,-1,-1,
      -1,10,11,12,13,14,15,-1,-1,-1,-1,-1,-1,-1,-1,-1,
      -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
      -1,10,11,12,13,14,15,-1,-1,-1,-1,-1,-1,-1,-1,-1,
      -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
      -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
      -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
      -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
      -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
      -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
      -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
      -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
      -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1
    };
  for (i = c = 0; i<length; i++)
  {
    int d = hex_digits[digits[i]];
    if (d < 0)
      return 0;
    c = 16 * c + d;
  }
  *value = c;
  return 1;
}

static int
parse_wchar_t(const struct buffer *buffer, wchar_t *value)
{
  if (buffer->length < 4)
    return 0;
  return parse_hex(buffer->current, buffer->length, value);
}

static int
parse_string(const struct buffer *buffer, const unsigned char **value)
{
  unsigned char *s;  
  
  s = malloc(buffer->length + 1);
  if (!s)
    return 0;

  memcpy(s, buffer->current, buffer->length);
  s[buffer->length] = '\0';

  *value = s;

  return 1;
}

static int
parse_decimal(const struct buffer *buffer, int *value)
{
  unsigned i;
  int x;
  

  if (!buffer->length)
    return 0;
  
  for (x = i = 0; i<buffer->length; i++)
  {
    unsigned char d = buffer->current[i];
    if (d < '0' || d > '9')
      return 0;

    x = 10 * x + (d - '0');
  }

  *value = x;
  return 1;  
}

static int
parse_category(const struct buffer *buffer, enum unidata_category *value)
{
  if (buffer->length != 2)
    return 0;

#define CAT(c) (!memcmp(buffer->current, #c, 2)) *value = UNIDATA_##c
  if CAT(Lu);
  else if CAT(Ll);
  else if CAT(Lt);
  else if CAT(Lm);
  else if CAT(Lo);
  else if CAT(Mn);
  else if CAT(Mc);
  else if CAT(Me);
  else if CAT(Nd);
  else if CAT(Nl);
  else if CAT(No);
  else if CAT(Pc);
  else if CAT(Pd);
  else if CAT(Ps);
  else if CAT(Pe);
  else if CAT(Pi);
  else if CAT(Pf);
  else if CAT(Po);
  else if CAT(Sm);
  else if CAT(Sc);
  else if CAT(Sk);
  else if CAT(So);
  else if CAT(Zs);
  else if CAT(Zl);
  else if CAT(Zp);
  else if CAT(Cc);
  else if CAT(Cf);
  else if CAT(Cs);
  else if CAT(Co);
  else if CAT(Cn);
  else return 0;
#undef CAT
  return 1;
}

static int
has_suffix(const unsigned char *x, const unsigned char *suffix)
{
  size_t x_length = strlen(x);
  size_t suffix_length = strlen(suffix);

  return (x_length >= suffix_length
	  && !strcmp(x + (x_length - suffix_length), suffix));
}

int
unidata_parse_entry(struct unidata_entry *entry,
		    const unsigned char *line)
{
  struct buffer buffer;
  buffer_init(&buffer, line);

  if (buffer_next(&buffer) && parse_wchar_t(&buffer, &entry->code)
      && buffer_next(&buffer) && parse_string(&buffer, &entry->name)
      && buffer_next(&buffer) && parse_category(&buffer, &entry->category)
      && (buffer_next(&buffer) && buffer.length
	  && parse_decimal(&buffer, &entry->combining_class)))
  {
    entry->start = entry->name[0] == '<' && has_suffix(entry->name, ", First>");
    entry->end = entry->name[0] == '<' && has_suffix(entry->name, ", Last>");

    return 1;
  }
  else
    return 0;
}
